/*
 *	getdrv.c
 *
 *	MSDOS Disk BPB Checker
 */

#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/file.h>

#define	DOSDIRENT	32
#define	SECTSIZE	{512, 1024, 256, 128}
#define	MAX12BIT	(0xff0 - 0x002)
#define	byte2word(p)	((p)[0] + ((unsigned short)(p)[1] << 8))
#define	byte2dword(p)	((p)[0] + ((unsigned long)(p)[1] << 8) \
				+ ((unsigned long)(p)[2] << 16) \
				+ ((unsigned long)(p)[3] << 24))
#define	FS_FAT		"FAT     "
#define	FS_FAT12	"FAT12   "
#define	FS_FAT16	"FAT16   "
#define	FS_FAT32	"FAT32   "

#ifndef	L_SET
# ifdef	SEEK_SET
# define	L_SET	SEEK_SET
# else
# define	L_SET	0
# endif
#endif
#ifndef	O_BINARY
#define	O_BINARY	0
#endif

#ifndef	__attribute__
#define	__attribute__(x)
#endif

typedef struct _bpb_t {
	unsigned char jmpcode[3] __attribute__ ((packed));
	unsigned char makername[8] __attribute__ ((packed));
	unsigned char sectsize[2] __attribute__ ((packed));
	unsigned char clustsize __attribute__ ((packed));
	unsigned char bootsect[2] __attribute__ ((packed));
	unsigned char nfat __attribute__ ((packed));
	unsigned char maxdir[2] __attribute__ ((packed));
	unsigned char total[2] __attribute__ ((packed));
	unsigned char media __attribute__ ((packed));
	unsigned char fatsize[2] __attribute__ ((packed));
	unsigned char secttrack[2] __attribute__ ((packed));
	unsigned char nhead[2] __attribute__ ((packed));
	unsigned char hidden[4] __attribute__ ((packed));
	unsigned char bigtotal[4] __attribute__ ((packed));
	unsigned char bigfatsize[4] __attribute__ ((packed));
	unsigned char exflags[2] __attribute__ ((packed));
	unsigned char filesys[2] __attribute__ ((packed));
	unsigned char rootdir[4] __attribute__ ((packed));
	unsigned char fsinfo[4] __attribute__ ((packed));
	unsigned char reserved[2] __attribute__ ((packed));
	unsigned char fsname[8] __attribute__ ((packed));
} bpb_t;

static int readbpb();

static short sectsizelist[] = SECTSIZE;


static int readbpb(dev)
char *dev;
{
	long total, nsect, nhead, ncyl, totalsize;
	long clustsize, sectsize, fatofs, fatsize, dirofs, dirsize, rootdir;
	bpb_t *bpb;
	char *buf;
	int i, cc, fd, fattype;

	if ((fd = open(dev, O_RDONLY | O_BINARY, 0600)) < 0) return(-1);
	if (lseek(fd, 0L, L_SET) < 0) {
		close(fd);
		return(-1);
	}

	cc = 0;
	for (i = 0; i < (sizeof(sectsizelist) / sizeof(short)); i++)
		if (cc < sectsizelist[i]) cc = sectsizelist[i];
	if (!(buf = (char *)malloc(cc))) {
		close(fd);
		return(-1);
	}
	for (i = 0; i < (sizeof(sectsizelist) / sizeof(short)); i++) {
		while ((cc = read(fd, buf, sectsizelist[i])) < 0
		&& errno == EINTR);
		if (cc >= 0) break;
	}
	if (i >= (sizeof(sectsizelist) / sizeof(short))) {
		close(fd);
		free(buf);
		return(-1);
	}
	bpb = (bpb_t *)buf;

	total = byte2word(bpb -> total);
	if (!total) total = byte2dword(bpb -> bigtotal);
	nsect = byte2word(bpb -> secttrack);
	nhead = byte2word(bpb -> nhead);
	ncyl = total / (nhead * nsect);

	clustsize = bpb -> clustsize;
	sectsize = byte2word(bpb -> sectsize);
	fatofs = byte2word(bpb -> bootsect);
	fatsize = byte2word(bpb -> fatsize);
	dirsize = (long)byte2word(bpb -> maxdir) * DOSDIRENT / sectsize;

	fattype = 0;
	if (fatsize && dirsize) rootdir = 0;
	else {
		if (!fatsize) fatsize = byte2dword(bpb -> bigfatsize);
		dirsize = 0;
		rootdir = byte2dword(bpb -> rootdir);
		fattype = 32;
	}

	dirofs = fatofs + fatsize * (bpb -> nfat);
	total -= dirofs + dirsize;
	totalsize = total / clustsize;

	if (!fattype) {
		if (!strncmp(bpb -> fsname, FS_FAT, sizeof(bpb -> fsname))
		|| !strncmp(bpb -> fsname, FS_FAT12, sizeof(bpb -> fsname)))
			fattype = 12;
		else if (!strncmp(bpb -> fsname, FS_FAT16,
		sizeof(bpb -> fsname))
		|| totalsize > MAX12BIT)
			fattype = 16;
		else fattype = 12;
	}

	printf("setdrv A \"%s\" %d, %d, %d\n\n",
		dev, nhead, nsect, ncyl);

	printf("Appendix:\n");
	printf("  cluster           = %d sectors\n", clustsize);
	printf("  sector            = %d bytes\n", sectsize);
	printf("  FAT offset        = %d th sector\n", fatofs);
	printf("  FAT size          = %d sectors\n", fatsize);
	printf("  FAT type          = %d bit\n", fattype);
	if (fattype == 32) {
		printf("  dir. entry offset = %d th cluster\n", rootdir);
		printf("  dir. entry size   = infinity\n");
	}
	else {
		printf("  dir. entry offset = %d th sector\n", dirofs);
		printf("  dir. entry size   = %d sectors\n", dirsize);
	}
	printf("  media descriptor  = 0x%02x\n", bpb -> media);

	free(buf);
	return(0);
}

int main(argc, argv)
int argc;
char *argv[];
{
	if (argc != 2 || argv[1][0] != '/') {
		fprintf(stderr, "usage: getdrv <device_file>\n");
		return(1);
	}
	if (readbpb(argv[1]) < 0) {
		fprintf(stderr, "%s: Cannot read.\n", argv[1]);
		return(1);
	}
	return(0);
}
